/// AudioDevice.h
///
/// (c) 2001-2006 by Bastian Spiegel <bs@tkscript.de>
///     - Distributed under terms of the Lesser GNU General Public License (LGPL).
///       See COPYING and <http://www.gnu.org/licenses/licenses.html#LGPL> for further information.
///

#ifndef __TKSDL_AUDIODEVICE_H__
#define __TKSDL_AUDIODEVICE_H__



#define TKS_AMIX_BUFSIZE 8192*2

YG("audio");

YF sF32 YAC_CALL _MIDINoteToFrequency(sF32);


YCS class _AudioDevice : public YAC_Object {
public:
   static sSI cbid_SDL_FillAudioBuffer;

public:

    SDL_AudioSpec *dsp_spec;

    YAC_String    filename; 
    void         *ofs;
    sBool         b_file;
    sBool         b_enabled;
    sUI           def_buf_size;

    sF32  volume;
    sBool swap_stereo;
    sF32  bpm; // beats per minute
    sUI   ppq; // precision per quarter

    sF32  dsp_frequency;
    sU16  dsp_block_samples;
    sUI   dsp_block_size;
    sU8   dsp_bits;     
    sU8   dsp_channels;
    sBool dsp_active;
    sBool running;
    
    // number of samples to process in current time frame
    // calculated by bpm, ppq and dsp_frequency
    sU32 timeframe_size; 

    // number of samples calculated for current time frame
    // if timeframe_samples_done==timeframe_size the mixer settings have to be updated (sequencer)
    sUI  timeframe_samples_done;

    sUI  timeframe_next_offset;
    sUI  timeframe_last_offset;
    sF32 timeframe_real_size;
    sU16 timeframe_count;

    sF32 mix_buffer[TKS_AMIX_BUFSIZE];
    sS16*dsp_buffer;

    sUI  req_size; // number of stereo samples to process during buffer fill
    sUI  req_samples_done; /// number of samples processed for current buffer fill

    void *sig_time;
    void *sig_postprocess;
    void *sig_render;
    void *sig_block;

    void*mtx_beginblock;
    
    void*sc_audiocontext;

    // anonymous pointer to FloatArray for onAudioPostProcess callback
    void *tmp_fa;

public:
    _AudioDevice(void);
    ~_AudioDevice();

    YAC(_AudioDevice);

    void YAC_CALL yacGetSignalStringList(YAC_String *);
    void YAC_CALL yacRegisterSignal(sUI _sigid, void *_fun);

    void beginBlock(sU32 _req_size, sS16 *_stereobuf16);
    void calcTimeframeSize(void);
    void beginTimeframe(void);
    void buildFrequencyTable(void);

    void YAC_CALL yacOperator(sSI, YAC_Object *, YAC_Value *);

    YM void _setVolume            (sF32);
    YM sF32 _getVolume            (void);
    YM void _setSwapstereo        (sSI);
    YM sSI  _getSwapstereo        (void);
    YM sSI  _openDSP              (sF32 _freq, sSI _bufsize);
    YM void _start                (void);
    YM void _stop                 (void);
    YM void _closeDSP             (void);
    YM sF32 _getBpm               (void);
    YM void _setBpm               (sF32);
    YM void _setPpq               (sSI _ppq);
    YM sSI  _getPpq               (void);
    YM void _processTimeFrames    (void);
    YM void _finishBlock          (void);
    YM void _setFileName          (YAC_String *_s);
    YM void _enableFileOutput     (sSI);
    YM void _enable               (sSI);
    YM void _setDefaultBufferSize (sSI);
};



#endif
