/// tkshal.h
///
/// (c) 2001-2005 by Bastian Spiegel <bs@tkscript.de>
///     - Distributed under terms of the Lesser GNU General Public License (LGPL).
///       See COPYING and <http://www.gnu.org/licenses/licenses.html#LGPL> for further information.
///
///
#ifndef __TKSDL_HAL_H__
#define __TKSDL_HAL_H__

YG("hal");

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//~                                                                   ~
//~ HAL Hardware Abstraction Layer using SDL, the Simple Direct Layer ~
//~                                                                   ~
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// ---- "scrap" is actually an example code by SDL author Sam Lantinga which demonstrates
// ---- how to access the X11/Win32/QNX clipboard. 
// ---- define this constant if you want to enable "scrap" support.

#ifdef YAC_WIN32
   #define TKSDL_WIN32_CLIPBOARD defined
   ////#define TKSDL_SCRAP_CLIPBOARD defined
#else
  #ifdef YAC_MACOSX
     #define TKSDL_MACOSX_CLIPBOARD defined
  #else
     #ifdef YAC_LINUX
        #define TKSDL_SCRAP_CLIPBOARD defined
     #else
        #define TKSDL_STUB_CLIPBOARD defined
     #endif
  #endif
#endif


// ---- assumes sames element alignment like in tkopengl plugin (usually 8 bytes) ----
class sFVector : public YAC_Object { public: sF32 x,y,z; };
class _Vector  : public sFVector   { };
class _IVector : public YAC_Object { public: sSI x; sSI y; };

extern sUI GetClassID_IVector(void);
extern sUI GetClassID_Vector(void);

enum __tksapplicationsignalids {
	TKS_SIG_HALSIGNAL=0,
	TKS_SIG_HALKEYBOARD,
	TKS_SIG_HALMOUSE,
	TKS_SIG_HALJOYAXISMOTION,
	TKS_SIG_HALJOYBUTTON,
	TKS_SIG_HALJOYHATMOTION,
	TKS_SIG_HALJOYBALLMOTION,
	TKS_SIG_HALDRAW,
	TKS_SIG_HALTICK,
	TKS_SIG_HALEXIT,
	TKS_SIG_HALUSEREVENT,
	TKS_SIG_HALREOPEN,
	TKS_SIG_HALRESIZE,
   TKS_SIG_HALEXPOSE,
   TKS_SIG_HALAPPLICATIONACTIVE,
   TKS_SIG_HALMOUSEFOCUSACTIVE,
   TKS_SIG_HALIDLE,

	TKS_HAL_NUMSIGNALS
};


enum __tksmouseflags {
    TKS_MOUSE_LBUTTON     =(1<<0),
    TKS_MOUSE_RBUTTON     =(1<<1),
    TKS_MOUSE_MBUTTON     =(1<<2),
    TKS_MOUSE_WHEELUP     =(1<<3),
    TKS_MOUSE_WHEELDOWN   =(1<<4),

    TKS_MOUSE_PAD
};

enum __mousepointermodes {
	TKS_MOUSE_POINTER_ENTER  =0,
	TKS_MOUSE_POINTER_MOVE   =1,
	TKS_MOUSE_POINTER_MANUAL =2
};

enum __tkspriorities {
    TKS_PRIORITY_IDLE=1,
    TKS_PRIORITY_NORMAL=2,
    TKS_PRIORITY_HIGH=3,
    TKS_PRIORITY_REALTIME=4
};

enum __tksredrawflags {
    TKS_REDRAW_ALL=0,
    TKS_REDRAW_DIRTY=1,

    TKS_REDRAW_PAD
};

enum __tkssdlevents {
	TKS_EVENT_ONACTION=0,

	TKS_EVENT_PAD
};

#define _HAL _SDL

YCS class _SDL : public YAC_Object {
public:

//#define MOUSE_LBUTTON                    YCI 1
//#define MOUSE_RBUTTON                    YCI 2
//#define MOUSE_MBUTTON                    YCI 4
//#define MOUSE_WHEELUP                    YCI 8
//#define MOUSE_WHEELDOWN                  YCI 16
//#define MOUSE_POINTER_ENTER              YCI 0
//#define MOUSE_POINTER_MOVE               YCI 1
//#define MOUSE_POINTER_MANUAL             YCI 2
//#define PRIORITY_IDLE                    YCI 1
//#define PRIORITY_NORMAL                  YCI 2
//#define PRIORITY_HIGH                    YCI 3
//#define PRIORITY_REALTIME                YCI 4
//#define REDRAW_ALL                       YCI 0
//#define REDRAW_DIRTY                     YCI 1
//#define SIGINT                           YCI 2
//#define SIGILL                           YCI 4
//#define SIGFPE                           YCI 8
//#define SIGSEGV                          YCI 11
//#define SIGTERM                          YCI 15
//#define SIGABRT                          YCI 22
//#define DEFAULT_KEYBOARD_REPEAT_DELAY    YCI 500
//#define DEFAULT_KEYBOARD_REPEAT_INTERVAL YCI 30
//#define SDL_RESIZABLE                    YCI 0x10
//#define SDL_NOFRAME                      YCI 0x20


   sBool b_unicode;

	sBool        b_running;
	sBool        b_graphics_active;
	sBool        b_event_polling;
	sBool        b_fullscreen;
	sBool        b_reopen;
	sBool        b_need_redraw;
	sBool        b_redrawing;
	sU8          redraw_flag;
	sU32         tks_flags; // viewport flags..
	void        *signal_funs[TKS_HAL_NUMSIGNALS];

   sBool b_double_buffer; // 1=create double buffered GL context
    
	YAC_String   window_title;
	sU16         view_flags;
	sU16         view_sx;
	sU16         view_sy;
	sU8          view_z;
	sU16         win_sx;
	sU16         win_sy;
	sU16         fs_sx;
	sU16         fs_sy;
	sU16         fs_z;

    sU32         now;
    sU32         last_time;
    struct _fps {
      sF32 tick_interval; // default interval of onTick(), also affects fps.precision (onDraw)
      sU32 tick_buffer;   // max. number of tick() calls in a row
      sU32 num_frames_rendered;
      sU32 total_time;
      sU32 second_time;
      sU32 last_secondtime;
      sU32 second_frames_rendered;
      sF32 real_fps;
      sU32 last_frametime;
      sF32 average_fps;
      sF32 current_fps;
      sF32 precision;
      sUI  limit_fps; // ---- limit redraw frequency to e.g. 25fps, 0=no limit ----
    } fps;
    sU32 max_frame_deltatime;

    struct _mouse {
      sS16                last_absx;
      sS16                last_absy;
      sS16                absx;
      sS16                absy;
      sS16                dx;
      sS16                dy;
      sU16                cstate;
      sU16                nstate;
      sBool               grabbed;
      sU16                warp_sx;
      sU16                warp_sy;
      sU8                 show;
      sU8                 pointer_mode;
      sU32                auto_hide_timeout;
      sU32                def_auto_hide_timeout;
      
    } mouse;

    struct _joysticks {
      sBool enabled;
      sS32  num;
      void *elements; // TKSJoystick object array
    } joysticks;
    
    struct __pickcolors {
      sU8 gb[2*32*32]; // 0=unused, 16 bit 565, 555 rgb, only 55 gb is used to be safe->1024-1 objects
      sBool first_tag;
      sU16 posx;
      sU16 posy;
      sU8  rgb[3];
      sF32 z;
      sF32 retz;
    } picking;

    enum __tksviewportflags {
      VP_OPENED    = (1<<0),
      VP_FULLSCREEN= (1<<1),
      
      VP_PAD
    };
    
    struct __audio {
      sF32 dsp_frequency;
    } audio;
    
#ifdef YAC_WIN32
    DWORD dw_monitorfreq;
    LPDIRECTDRAW lpDD;
#endif


#ifdef TKSDL_SCRAP_CLIPBOARD
    YAC_String scrap_string;
#endif
	

public:
	 _HAL(void);
	 ~_HAL();

	 YAC(_HAL);

	 void  YAC_CALL yacGetSignalStringList(YAC_String *);
	 void  YAC_CALL yacRegisterSignal(sUI _sigid, void *_fun);

	 void  setArgs(int _argc, char **_argv);
	 void  tick(void);
	 void  draw(void);
	 sBool loopDelay(void);
	 void  showCursor(sBool);
	 void  beginPaint(void);
	 void  beginPaintDirty(void);
	 void  endPaint(void);
	 void  callExit(void);
	 void  callReopen(void);
 	 void  callResize(void);
 	 void  callExpose(void);
    void  callApplicationActive(sSI _gain);
    void  callMouseFocusActive(sSI _gain);
	 void  callOnJoyAxisMotion(sS32,sS32,sS32);
	 void  callOnJoyButton(sS32,sS32,sS32);
	 void  callOnJoyHatMotion(sS32,sS32,sS32);
	 void  callOnJoyBallMotion(sS32,sS32,sS32,sS32);
	 void  initFPS(void);
	 sBool processSDLEvent(sBool _poll);

	 // ---- keyboard input
	 void  callKeyboardHandler(sUI _vkey, sUI _vmod, sUI _unicode);

	 void  callMouseHandler(sS32 _mx, sS32 _my, sS32 _cstate, sS32 _nstate);
	 void  resetMouse(void);

	 // ---- joystick input
	 void  openJoysticks(void);
	 void  closeJoysticks(void);

	 // ---- viewport
	 sSI   openView(sU16 _sx, sU16 _sy, sU8 _z, sBool _fullscreen, sBool _resize);
	 sBool openWindow(sS32 _sx, sS32 _sy);
	 sBool openScreen(sS32 _sx, sS32 _sy, sS32 _z);
	 void  viewOpened(void);
	 void  closeView(void);
	 void  useWindow(sU16 _sx, sU16 _sy);
	 sBool toggleFullScreen(void);

	 // ---- vertical sync
	 void  waitVBlank(void);
	 sUI   getMonitorFrequency(void);

	 // ---- 3D picking support
	 sU8  *allocPickingColor(void);
	 sF32  getPickedZ(void);
	 sUI   getPickedID(void);

    // ---- unicode support
    void updateUnicode(void);

public:
	 // ------------ exported methods ------------
	 YM void _eventLoop            (void);
	 YM void _exitEventLoop        (void);
	 YM void _setPriority          (sSI _prio);
	 YM sSI  _getTicks             (void);
	 YM void _triggerEvent         (YAC_Object *);
	 YM void _setRedrawFlag        (sSI);
	 YM sSI  _getRedrawFlag        (void);
	 YM void _delay                (sSI);
	 YM void _sendUserEvent        (YAC_Object *_s);
	 YM void _enableKeyRepeat      (sSI _delay, sSI _rate);
	 YM sSI  _getNumJoysticks      (void);
	 YM void _getJoystick          (sSI, YAC_Value*);
	 YM void _updateJoysticks      (void);
	 YM void _reportJoystickEvents (sSI);
	 YM void _setEventPolling      (sSI);
	 YM sSI  _getEventPolling      (void);
	 YM sUI  _allocPickingColor    (void);
	 YM void _initPickingColors    (void);
    YM sSI  _enableUNICODE        (sSI _enabled);
    YM sSI  _isApplicationActive  (void);
    YM sSI  _getSysWMInfo         (void);
};	 

extern _HAL *tks_HAL;

#endif
