/// Mouse.cpp
///
/// (c) 2001-2005 by Bastian Spiegel <bs@tkscript.de>
///     - Distributed under terms of the Lesser GNU General Public License (LGPL).
///       See COPYING and <http://www.gnu.org/licenses/licenses.html#LGPL> for further information.
///

#include "static_tksdl.h"

#include <stdlib.h>

#define YAC_PRINTF defined
#include <yac.h>

#ifdef YAC_WIN32
#include <windows.h>
#include <ddraw.h>
#endif
#include "HAL.h"

#include "Mouse.h"

extern YAC_Host *yac_host;
extern sUI clid_X_IVector;
extern sUI clid_X_Vector;
extern sUI clid_Joystick;

#include <SDL/SDL.h>
#include <GL/glu.h>


void _Mouse::_grab(void) {
   ::SDL_WM_GrabInput(SDL_GRAB_ON);
   ::SDL_WarpMouse(tks_HAL->view_sx/2, tks_HAL->view_sy/2);
   tks_HAL->mouse.grabbed=1;
   tks_HAL->mouse.show=TKS_MOUSE_SHOW_DISABLE;
}

void _Mouse::_ungrab(void) {
   tks_HAL->mouse.grabbed=0;
   tks_HAL->mouse.show=TKS_MOUSE_SHOW_ENABLE;
   ::SDL_WM_GrabInput(SDL_GRAB_OFF);
   tks_HAL->mouse.warp_sx=tks_HAL->mouse.absx;
   tks_HAL->mouse.warp_sy=tks_HAL->mouse.absy;
}

void _Mouse::_toggleGrab() {
   if(tks_HAL->mouse.grabbed)
      _ungrab();
   else
      _grab();
}

sSI _Mouse::_getX(void) {
   return tks_HAL->mouse.absx;
}

sSI _Mouse::_getY(void) {
   return tks_HAL->mouse.absy;
}

sSI _Mouse::_getDx(void) {
   return tks_HAL->mouse.dx;
}

sSI _Mouse::_getDy(void) {
   return tks_HAL->mouse.dy;
}

sSI _Mouse::_getButtons(void) {
   return tks_HAL->mouse.cstate;
}

sSI _Mouse::_getChangedButtons(void) {
   return tks_HAL->mouse.nstate;
}

sSI _Mouse::_isGrabbed(void) {
   return tks_HAL->mouse.grabbed;
}


void _Mouse::_showPointer(sSI _enabled) {
   tks_HAL->showCursor(_enabled);
}

void _Mouse::_unproject(YAC_Object *_rnear, YAC_Object *_rfar) {
   if(YAC_CHK(_rnear, GetClassID_Vector()))
      if(YAC_CHK(_rfar, GetClassID_Vector()))
      {
         GLdouble mvmatrix[4*4];
         GLdouble prmatrix[4*4];
         glGetDoublev(GL_MODELVIEW_MATRIX, mvmatrix);
         glGetDoublev(GL_PROJECTION_MATRIX, prmatrix);
         _Vector *vnear=(_Vector*)_rnear;
         _Vector *vfar=(_Vector*)_rfar;
         GLdouble x=(GLdouble)tks_HAL->mouse.absx;
         GLdouble y=(GLdouble)tks_HAL->view_sy-tks_HAL->mouse.absy; //tks_HAL->mouse.last_ev_motion_y;
         GLdouble z=0.0;
         GLint view[4];
         glGetIntegerv(GL_VIEWPORT, view);
         GLdouble rx,ry,rz;
         ::gluUnProject(x,y,z, mvmatrix, prmatrix, view, &rx,&ry,&rz);
         vnear->x=(sF32)rx;
         vnear->y=(sF32)ry;
         vnear->z=(sF32)rz;
         z=1.0;
         ::gluUnProject(x,y,z, mvmatrix, prmatrix, view, &rx,&ry,&rz);
         vfar->x=(sF32)rx;
         vfar->y=(sF32)ry;
         vfar->z=(sF32)rz;
      }
}

void _Mouse::_setPointerMode(sSI _m) {
   switch(_m)
   {
   case TKS_MOUSE_POINTER_ENTER:
   case TKS_MOUSE_POINTER_MOVE:
      tks_HAL->mouse.auto_hide_timeout=tks_HAL->mouse.def_auto_hide_timeout;
      tks_HAL->mouse.pointer_mode=_m;
      break;
   case TKS_MOUSE_POINTER_MANUAL:
      tks_HAL->mouse.auto_hide_timeout=0;
      tks_HAL->mouse.pointer_mode=TKS_MOUSE_POINTER_MANUAL;
      break;
   }
}

sSI _Mouse::_getPointerMode(void) {
   return tks_HAL->mouse.pointer_mode;
}
