/// Thread.cpp
///
/// (c) 2002-2005 Bastian Spiegel <bs@tkscript.de> 
///     - Distributed under terms of the Lesser GNU General Public License (LGPL).
///       See COPYING and <http://www.gnu.org/licenses/licenses.html#LGPL> for further information.
///
///
/// changelog:
///      04Jul2002 <fli> file created
///      11Mar2005 source cleanup
///

#include "static_tksdl.h"

#include <stdio.h>
#include <stdlib.h>

#include <SDL/SDL.h>
#include <SDL/SDL_thread.h>

#include <yac.h>

#ifndef DX_STATIC_TKSDL
#include <tks-list.h>
#endif

#include <tks-opengl.h>
#include "Thread.h"

extern YAC_Host *yac_host;

int TKS_SDL_Thread_Entry(void *_dat) {
	if(_dat)
	{
		_Thread *dat=(_Thread *)_dat;
		YAC_Value args[1];
		args[0].initObject(dat, 0);
      void *ctx = yac_host->yacContextCreate();
		yac_host->yacEvalFunction(ctx, dat->tks_entry_fun, 1, args);
      yac_host->yacContextDestroy(ctx);
		dat->sdl_thread=0;
	}
	return 0;
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//~                                   ~
//~ Thread                            ~
//~                                   ~
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
_Thread::_Thread(void) {
  sdl_thread=0;
  user_data=0;
  tks_entry_fun=0;
  mtx_event_queue=0;
  action_queue=0;
  //StaticList::InsertHead(&action_queue, 0);
}

_Thread::~_Thread() {
  _cancelEvents();
  _kill();
}

sSI _Thread::_create(YAC_Object *_fun) {
  if((!sdl_thread) && YAC_CHK(_fun, YAC_CLID_STRING))
    {
      if( (tks_entry_fun=yac_host->yacFindFunction((sChar*)((YAC_String*)_fun)->chars)) )
	{
	  mtx_event_queue=::SDL_CreateMutex();
	    if(mtx_event_queue)
	      {
		mtx_event_wait=::SDL_CreateMutex();
		if(mtx_event_wait)
		  {
		    SDL_mutexP(mtx_event_wait);
		    sdl_thread=::SDL_CreateThread(TKS_SDL_Thread_Entry, (void*)this);
		    if(sdl_thread)
		      return 1;
		    else
		      yac_host->printf("[---] SDL_CreateThread() failed.\n");
		  }
		else
		  yac_host->printf("[---] failed to create Mutex mtx_event_wait.\n");
	      }
	    else
	      yac_host->printf("[---] failed to create Mutex mtx_event_queue.\n");
	}
      else
	yac_host->printf("[---] Thread::create: function \"%s\" not found (Module.FunctionName syntax!).\n",
			 (char*)((YAC_String*)_fun)->chars);
    }
  return 0;
}

void _Thread::_kill(void) {
	if(sdl_thread)
	{
		yac_host->printf("[---] killing Thread leftovers..\n");
		::SDL_KillThread(sdl_thread);
		sdl_thread=0;
	}
	if(mtx_event_queue)
	{
		SDL_DestroyMutex(mtx_event_queue);
		mtx_event_queue=0;
	}
	if(action_queue)
		StaticList::Free(&action_queue);
}

sSI _Thread::_getId(void) {
	if(sdl_thread)
		return ::SDL_GetThreadID(sdl_thread);
	return -1;
}

sSI  _Thread::_wait(void) {
	if(sdl_thread)
	{
		sSI ret;
		::SDL_WaitThread(sdl_thread, &ret);
		return ret;
	}
	return -1;
}

void _Thread::_event(YAC_Object *_s) {
	if(mtx_event_queue)
	{
		::SDL_mutexP(mtx_event_queue);
		if(YAC_CHK(_s, YAC_CLID_STRING))
		{
			YAC_Event *ac=(YAC_Event*)yac_host->yacNewByID(YAC_CLID_EVENT);
			ac->time_stamp=::SDL_GetTicks();
			ac->data.yacCopy((YAC_String*)_s);
			StaticList::InsertHead(&action_queue, (sU32)ac);
		}
		else
			if(YAC_CHK(_s, YAC_CLID_EVENT))
			{
				YAC_Event *o=(YAC_Event*)_s;
				YAC_Event *ac=(YAC_Event*)yac_host->yacNewByID(YAC_CLID_EVENT);
				ac->time_stamp=o->time_stamp;
				ac->data.yacCopy(&o->data);
				StaticList::InsertHead(&action_queue, (sU32)ac);
			}
		::SDL_mutexV(mtx_event_queue);
		::SDL_mutexV(mtx_event_wait);  
	}
}

void _Thread::_cancelEvents(void) {
	if(mtx_event_queue)
		::SDL_mutexP(mtx_event_queue);
	StaticList::Node *c=action_queue;
	while(c)
	{
		YAC_Event *ac=(YAC_Event*)c->data;
		if(ac)
		{
			delete ac;
			c->data=0;
		}
		c=c->previous;
	}
	if(mtx_event_queue)
		::SDL_mutexV(mtx_event_queue);
}

void _Thread::_setUserdata(YAC_Object *_o) {
	user_data=_o;
}

YAC_Object *_Thread::_getUserdata(void) {
	return user_data;
}

void _Thread::_getNextEvent(YAC_Value *_r) {
	if(mtx_event_queue)
		::SDL_mutexP(mtx_event_queue);
	StaticList::Node *l=StaticList::Tail(action_queue);
	if(l)
	{
		if(l->data)
		{
			YAC_RETO((YAC_Event*)l->data, 1);
			StaticList::Remove(l);
			if(l==action_queue)
				action_queue=0;
			if(mtx_event_queue)
				SDL_mutexV(mtx_event_queue);
			return;
		}
	}
	YAC_RETO(0, 0);
	if(mtx_event_queue)
		SDL_mutexV(mtx_event_queue);
}

void _Thread::_waitForNextEvent(YAC_Value *_r) {
	::SDL_mutexP(mtx_event_wait); // try to lock mutex, wait until it's unlocked
	_getNextEvent(_r);
	::SDL_mutexP(mtx_event_queue);
	if(action_queue)
		::SDL_mutexV(mtx_event_wait);
	::SDL_mutexV(mtx_event_queue);
}
