/// Viewport.cpp
///
/// (c) 2001-2006 by Bastian Spiegel <bs@tkscript.de>
///     - Distributed under terms of the Lesser GNU General Public License (LGPL).
///       See COPYING and <http://www.gnu.org/licenses/licenses.html#LGPL> for further information.
///

#include "static_tksdl.h"

#include <stdlib.h>

#define YAC_PRINTF defined
#include <yac.h>

#ifdef YAC_WIN32
#include <windows.h>
#include <ddraw.h>
#endif
#include "HAL.h"

#include "Viewport.h"
#include "Mouse.h"

extern YAC_Host *yac_host;
extern sUI clid_X_IVector;
extern sUI clid_X_Vector;
extern sUI clid_Joystick;

#include <SDL/SDL.h>


_Viewport::_Viewport(void) {
}

_Viewport::~_Viewport() {
	tks_HAL->closeView();
}

void _Viewport::_toggleFullScreen(void) {
	tks_HAL->toggleFullScreen();
}

sSI _Viewport::_openWindow(sSI _sx, sSI _sy) {
	return tks_HAL->openWindow(_sx,_sy);
}

sSI _Viewport::_resizeWindow(sSI _sx, sSI _sy) {
	if(!(tks_HAL->view_flags&_HAL::VP_FULLSCREEN))
		if(tks_HAL->view_flags&_HAL::VP_OPENED)
			if((_sx>0) && (_sy>0))
         {
				return tks_HAL->openView(_sx, _sy, 0,0, 1);
         }
	return 0;
}

void _Viewport::_setFlags(sSI _fl) {
	tks_HAL->tks_flags=_fl&(SDL_RESIZABLE|SDL_NOFRAME);
}

sSI _Viewport::_getFlags(void) {
	return tks_HAL->tks_flags;
}

sSI _Viewport::_openScreen(sSI _sx, sSI _sy, sSI _z) {
	return tks_HAL->openScreen(_sx, _sy, _z);
}

void _Viewport::_close(void) {
    tks_HAL->closeView();
}

void _Viewport::_waitVBlank(void) {
	tks_HAL->waitVBlank();
}

sSI _Viewport::_getFrequency(void) {
	return tks_HAL->getMonitorFrequency();
}

sF32 _Viewport::_getPickedz(void) {
	return tks_HAL->getPickedZ();
}

sSI _Viewport::_getPickedid(void) {
	return tks_HAL->getPickedID();
}

void _Viewport::_setPickingPos(sSI _x, sSI _y) {
    tks_HAL->picking.posx=(sU16)_x;
    tks_HAL->picking.posy=tks_HAL->view_sy-(sU16)_y;
}

void _Viewport::_beginPaint(void) {
	tks_HAL->beginPaint();
}

void _Viewport::_endPaint(void) {
	tks_HAL->endPaint();
}

sSI _Viewport::_getWidth(void) {
	return tks_HAL->view_sx;
}

sSI _Viewport::_getHeight(void) {
	return tks_HAL->view_sy;
}

sSI _Viewport::_getDepth(void) {
	return tks_HAL->view_z;
}

void _Viewport::_setCaption(YAC_String *_sw) {
    ::SDL_WM_SetCaption((char*)_sw->chars, 0);
}

void _Viewport::_needRedraw(void) {
	tks_HAL->b_need_redraw=1;
}

void _Viewport::_setScreenResolution(sSI _sx, sSI _sy, sSI _bits) {
    if(_sx>0 && _sy>0 && _bits>0)
    {
        tks_HAL->fs_sx=_sx;
        tks_HAL->fs_sy=_sy;
        if(_bits!=16 && _bits!=32)
            if(_bits<16)
                _bits=16;
            else
                _bits=32;
        tks_HAL->fs_z=_bits;
    }
}

sSI  _Viewport::_getDoubleBuffer (void) {
   return tks_HAL->b_double_buffer;
}

void _Viewport::_setDoubleBuffer (sSI _enabled) {
   tks_HAL->b_double_buffer = (_enabled!=0);
}

