/// tksdl.cpp
///
/// (c) 2001-2006 by Bastian Spiegel <bs@tkscript.de>
///     - Distributed under terms of the Lesser GNU General Public License (LGPL).
///       See COPYING and <http://www.gnu.org/licenses/licenses.html#LGPL> for further information.
///
///

#include "static_tksdl.h"

#include <math.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define YAC_PRINTF defined
#define YAC_BIGSTRING defined
#include <yac.h>

#ifdef WIN32
#include <windows.h>
#endif

#include <SDL/SDL.h>
#include <SDL/SDL_net.h>
#include <SDL/SDL_thread.h>

#ifndef DX_STATIC_TKSDL
#include <tks-list.h>
#endif

#undef INADDR_ANY
#undef INADDR_NONE
#ifndef DX_STATIC_TKSDL
#include <yac_host.cpp>
#endif
#include <tks-opengl.h>
#include "IPAddress.h"
#include "TCPSocket.h"
#include "UDPPacket.h"
#include "UDPSocket.h"
#include "Network.h"
#include "HAL.h"
#include "Viewport.h"
#include "FPS.h"
#include "Mouse.h"
#include "Key.h"
#include "Joystick.h"
#include "Bitmap.h"
#include "Cursor.h"
#include "AudioDevice.h"
#include "AudioSample.h"
#include "AudioSampleVoice.h"
#include "SamplePlayer.h"
#include "Clipboard.h"


sUI clid_X_IVector;
sUI clid_X_Vector;
extern sUI clid_IPAddress;


_SDL     *tks_HAL;
#ifndef DX_STATIC_TKSDL
YAC_Host *yac_host;
#endif

#include "ying_hal.h"
#include "ying_hal1.h"
#include "ying_audio_AudioDevice.cpp"
#include "ying_audio_AudioSample.cpp"
#include "ying_audio_SamplePlayer.cpp"
#include "ying_audio.cpp"
#include "ying_hal1_Cursor.cpp"
#include "ying_hal1_FPS.cpp"
#include "ying_hal1_Joystick.cpp"
#include "ying_hal1_Key.cpp"
#include "ying_hal1_Mouse.cpp"
#include "ying_hal1_Viewport.cpp"
#include "ying_hal1.cpp"
#include "ying_hal_SDL.cpp"
#include "ying_hal.cpp"
#include "ying_network_IPAddress.cpp"
#include "ying_network_Network.cpp"
#include "ying_network_TCPSocket.cpp"
#include "ying_network_UDPPacket.cpp"
#include "ying_network_UDPSocket.cpp"
#include "ying_network.cpp"

void *sc_context;

// ---------------------------------------------------------------------------
#ifdef DX_STATIC_TKSDL
void YAC_CALL YAC_Init_tksdl(YAC_Host *_host) {
#else
void YAC_CALL YAC_Init(YAC_Host *_host) {
	yac_host=_host;
#endif

	if(_host->yacGetDebugLevel()) 
	  _host->printf("tksdl::YAC_Init called.\n");

   sc_context = yac_host->yacContextGetDefault();

#ifdef YAC_WIN32
	::SDL_SetModuleHandle(::GetModuleHandle(NULL)); 
#endif
	::SDL_Init(SDL_INIT_AUDIO | SDL_INIT_JOYSTICK | SDL_INIT_NOPARACHUTE);
	::SDLNet_Init();

	YAC_Init_hal(_host);
   tks_HAL=t_SDL->ctemplate;
   tks_HAL->b_unicode=::SDL_EnableUNICODE(-1)?1:0; // query SDL setting

	YAC_Init_hal1(_host);

	YAC_Init_audio(_host);

	YAC_Init_network(_host);

	clid_X_IVector = (sUI)-1;
	clid_X_Vector = (sUI)-1; // imported from tkopengl plugin

	tks_HAL->openJoysticks();

   _AudioDevice::cbid_SDL_FillAudioBuffer = yac_host->yacCallbackCreate("SDL_FillAudioBuffer");

	if(yac_host->yacGetDebugLevel())
		yac_host->yacPrint("tksdl::YAC_Init() finished.\n");
}

// the correct exit order is: 
//	delete t_FPS;
//	delete t_Joystick;
//	delete t_Key;
//	delete t_Mouse;
//	delete t_Viewport;
//	delete t_SDL;
// --- don't use *.h to create YAC bindings!

#ifdef DX_STATIC_TKSDL
void YAC_CALL YAC_Exit_tksdl(YAC_Host *_host) {
#else
void YAC_CALL YAC_Exit(YAC_Host *_host) {
#endif
   sBool bDebug=yac_host->yacGetDebugLevel()>0;
	if(bDebug)
		yac_host->yacPrint("tksdl::YAC_Exit() called.\n");

   if(bDebug)
      yac_host->yacPrint("sdl: exit network.\n");
   YAC_Exit_network      (_host);

   if(bDebug)
      yac_host->yacPrint("sdl: exit audio.\n");
   YAC_Exit_audio        (_host);

   if(bDebug)
      yac_host->yacPrint("sdl: exit hal1.\n");
   YAC_Exit_hal1         (_host);

   if(bDebug)
      yac_host->yacPrint("sdl: exit hal.\n");
   YAC_Exit_hal          (_host);

#ifndef DX_STATIC_TKSDL
	::SDLNet_Quit();
	::SDL_Quit();
#endif

	if(yac_host->yacGetDebugLevel())
		yac_host->yacPrint("tksdl::YAC_Exit() finished.\n");

}

#ifndef DX_STATIC_TKSDL
sUI YAC_CALL YAC_Version(void) {
	return 0x00090226;
}
#endif // not DX_STATIC_TKSDL

